# -*- coding: utf-8 -*-
"""Text parser plugin for XChat scrollback log files.

Information updated 06 September 2013.

Besides the logging capability, the XChat IRC client has the option to
record the text for opened tabs. So, when rejoining a particular channel
and/or a particular conversation, XChat will display the last messages
exchanged. This artifact could be present, if not disabled, even if
normal logging is disabled.

From the XChat FAQ (http://xchat.org/faq):

Q: 'How do I keep text from previous sessions from being displayed when I
join a channel?'
R: 'Starting in XChat 2.8.4, XChat implemented the Scrollback feature which
displays text from the last time you had a particular tab open.
To disable this setting for all channels, Go to Settings -> Preferences
-> Logging and uncheck Display scrollback from previous session.
In XChat 2.8.6, XChat implemented both Per Channel Logging, and
Per Channel Scrollbacks. If you are on 2.8.6 or newer, you can disable
loading scrollback for just one particular tab name by right clicking on
the tab name, selecting Settings, and then unchecking Reload scrollback'

The log file format differs from logging format, but it's quite simple
'T 1232315916 Python interface unloaded'
<T><space><decimal timestamp><space><text><\\n>

The time reported in the log is the number of seconds since January 1, 1970
00:00:00 UTC (from source code, time(0)). The <text> part could contain some
'decorators' (bold, underline, colors indication, etc.), so the parser
should strip those control fields.

Also see:
  http://xchat.org
"""

import pyparsing

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class XChatScrollbackEventData(events.EventData):
  """XChat Scrollback line event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the log entry
        was added.
    nickname (str): nickname.
    text (str): text sent by nickname service messages.
  """

  DATA_TYPE = 'xchat:scrollback:line'

  def __init__(self):
    """Initializes event data."""
    super(XChatScrollbackEventData, self).__init__(data_type=self.DATA_TYPE)
    self.added_time = None
    self.nickname = None
    self.text = None


class XChatScrollbackLogTextPlugin(interface.TextPlugin):
  """Text parser plugin for XChat scrollback log files."""

  NAME = 'xchatscrollback'
  DATA_FORMAT = 'XChat scrollback log file'

  ENCODING = 'utf-8'

  # Using a regular expression to include checking for spaces.
  _LOG_LINE = pyparsing.Regex(r'T (?P<timestamp>\d{1,10}) (?P<raw_text>.*)\n')

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE

  # Define for the stripping phase.
  _STRIPPER = (
      pyparsing.Word('\x03', pyparsing.nums, max=3).suppress() |
      pyparsing.Word('\x02\x07\x08\x0f\x16\x1d\x1f', exact=1).suppress())

  _NICKNAME = pyparsing.QuotedString('<', endQuoteChar='>').setResultsName(
      'nickname')

  # Define the structure for parsing <text> and get <nickname> and <text>
  _MESSAGE = (pyparsing.Optional(_NICKNAME) +
              pyparsing.restOfLine().setResultsName('text')).parseWithTabs()

  def _ParseRawText(self, raw_text):
    """Parses the raw text.

    This method implements the XChat strip_color2 and fe_print_text
    functions, slightly modified to get pure text. From the parsing point
    of view, after having stripped, the code takes everything as is,
    simply replacing tabs with spaces (as the original XChat code).

    Args:
      raw_text (str): raw text obtained from the log record.

    Returns:
      tuple[str, str]: nickname and text, or None if not available.

    Raises:
      ParseError: if the raw text cannot be parsed.
    """
    try:
      stripped_text = self._STRIPPER.transformString(raw_text)
      structure = self._MESSAGE.parseString(stripped_text)
    except pyparsing.ParseException as exception:
      raise errors.ParseError(
          'Unable to parse raw text with error: {0!s}'.format(exception))

    nickname = self._GetValueFromStructure(structure, 'nickname')
    text = self._GetValueFromStructure(structure, 'text', default_value='')
    text = text.replace('\t', ' ')
    return nickname, text

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    timestamp = self._GetValueFromStructure(structure, 'timestamp')
    timestamp = int(timestamp, 10)

    raw_text = self._GetValueFromStructure(
        structure, 'raw_text', default_value='')

    nickname, text = self._ParseRawText(raw_text)

    event_data = XChatScrollbackEventData()
    event_data.added_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
    event_data.nickname = nickname
    event_data.text = text

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    try:
      self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(XChatScrollbackLogTextPlugin)



# -*- coding: utf-8 -*-
"""The analysis plugin mediator object."""

import collections
import time

from plaso.containers import warnings
from plaso.engine import path_helper


class AnalysisMediator(object):
  """Analysis plugin mediator.

  Attributes:
    analysis_reports_counter (collections.Counter): number of analysis reports
        per analysis plugin.
    event_labels_counter (collections.Counter): number of event tags per label.
    last_activity_timestamp (int): timestamp received that indicates the last
        time activity was observed. The last activity timestamp is updated
        when the mediator produces an attribute container, such as an event
        tag. This timestamp is used by the multi processing worker process
        to indicate the last time the worker was known to be active. This
        information is then used by the foreman to detect workers that are
        not responding (stalled).
    number_of_produced_analysis_reports (int): number of produced analysis
        reports.
    number_of_produced_event_tags (int): number of produced event tags.
  """

  def __init__(self, data_location=None, user_accounts=None):
    """Initializes an analysis plugin mediator.

    Args:
      data_location (Optional[str]): location of data files used during
          analysis.
      user_accounts (Optional[list[UserAccountArtifact]]): user accounts.
    """
    super(AnalysisMediator, self).__init__()
    self._abort = False
    self._data_location = data_location
    self._event_filter_expression = None
    self._number_of_warnings = 0
    self._storage_writer = None
    self._user_accounts = user_accounts
    self._username_by_user_directory = {}

    self.analysis_reports_counter = collections.Counter()
    self.event_labels_counter = collections.Counter()
    self.last_activity_timestamp = 0.0
    self.number_of_produced_analysis_reports = 0
    self.number_of_produced_event_tags = 0

  @property
  def abort(self):
    """bool: True if the analysis should be aborted."""
    return self._abort

  @property
  def data_location(self):
    """str: path to the data files."""
    return self._data_location

  def GetDisplayNameForPathSpec(self, path_spec):
    """Retrieves the display name for a path specification.

    Args:
      path_spec (dfvfs.PathSpec): path specification.

    Returns:
      str: human readable version of the path specification.
    """
    return path_helper.PathHelper.GetDisplayNameForPathSpec(path_spec)

  def GetUsernameForPath(self, path):
    """Retrieves a username for a specific path.

    This is determining if a specific path is within a user's directory and
    returning the username of the user if so.

    Args:
      path (str): path.

    Returns:
      str: username or None if the path does not appear to be within a user's
          directory.
    """
    path = path.lower()

    username = self._username_by_user_directory.get(path, None)
    if not username and self._user_accounts:
      for user_account in self._user_accounts:
        if user_account.user_directory:
          user_directory = user_account.user_directory.lower()
          if path.startswith(user_directory):
            username = user_account.username
            self._username_by_user_directory[path] = username
            break

    return username

  def ProduceAnalysisResult(self, analysis_result):
    """Produces an analysis result attribute.

    Args:
      analysis_result (AttributeContainer): analysis result.
    """
    if self._storage_writer:
      self._storage_writer.AddAttributeContainer(analysis_result)

  def ProduceAnalysisReport(self, plugin):
    """Produces an analysis report.

    Args:
      plugin (AnalysisPlugin): plugin.
    """
    analysis_report = plugin.CompileReport(self)
    if not analysis_report:
      # TODO: produce AnalysisWarning that no report can be generated.
      return

    analysis_report.event_filter = self._event_filter_expression

    if self._storage_writer:
      self._storage_writer.AddAttributeContainer(analysis_report)

    self.analysis_reports_counter[analysis_report.plugin_name] += 1
    self.analysis_reports_counter['total'] += 1

    self.number_of_produced_analysis_reports += 1

    self.last_activity_timestamp = time.time()

  def ProduceAnalysisWarning(self, message, plugin_name):
    """Produces an analysis warning.

    Args:
      message (str): message of the warning.
      plugin_name (str): name of the analysis plugin to which the warning
          applies.
    """
    if self._storage_writer:
      warning = warnings.AnalysisWarning(
         message=message, plugin_name=plugin_name)
      self._storage_writer.AddAttributeContainer(warning)

    self._number_of_warnings += 1

    self.last_activity_timestamp = time.time()

  def ProduceEventTag(self, event_tag):
    """Produces an event tag.

    Args:
      event_tag (EventTag): event tag.
    """
    if self._storage_writer:
      self._storage_writer.AddOrUpdateEventTag(event_tag)

    for label in event_tag.labels:
      self.event_labels_counter[label] += 1
      self.event_labels_counter['total'] += 1

    self.number_of_produced_event_tags += 1

    self.last_activity_timestamp = time.time()

  def SetStorageWriter(self, storage_writer):
    """Sets the storage writer.

    Args:
      storage_writer (StorageWriter): storage writer.
    """
    self._storage_writer = storage_writer

  def SignalAbort(self):
    """Signals the analysis plugins to abort."""
    self._abort = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the MacOS Bluetooth plist plugin."""

import unittest

from plaso.parsers.plist_plugins import bluetooth

from tests.parsers.plist_plugins import test_lib


class MacOSBluetoothPlistPluginTest(test_lib.PlistPluginTestCase):
  """Tests for the MacOS Bluetooth plist plugin."""

  def testProcess(self):
    """Tests the Process function."""
    plugin = bluetooth.MacOSBluetoothPlistPlugin()
    storage_writer = self._ParsePlistFileWithPlugin(
        plugin, ['plist_binary'], 'com.apple.bluetooth.plist')

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 6)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'macos:bluetooth:entry',
        'device_identifier': '44-00-00-00-00-04',
        'device_name': 'Apple Magic Trackpad 2',
        'inquiry_time': '2012-11-02T01:13:17.324095+00:00',
        'is_paired': True,
        'name_update_time': '2012-11-02T01:21:38.997673+00:00',
        'services_update_time': '2012-11-02T01:13:23.000000+00:00'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the event attribute containers."""

import unittest

from plaso.containers import events

from tests import test_lib as shared_test_lib


class EventValuesHelperTest(shared_test_lib.BaseTestCase):
  """Tests for the event values helper functions."""

  def testCalculateEventValuesHash(self):
    """Tests the CalculateEventValuesHash function."""
    event_data = events.EventData()
    event_data.data_type = 'test'
    event_data.attribute1 = 'attribute1'
    event_data.attribute2 = 10
    event_data.attribute3 = ['attribute1']

    event_data_stream = events.EventDataStream()
    event_data_stream.attribute1 = 'ATTR1'
    event_data_stream.attribute2 = 99

    content_identifier = events.CalculateEventValuesHash(
        event_data, event_data_stream)

    self.assertEqual(content_identifier, '31aac7b1f8c1446f4b638c0dc5f92981')


class EventDataTest(shared_test_lib.BaseTestCase):
  """Tests for the event data attribute container."""

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = events.EventData()

    expected_attribute_names = [
        '_event_data_stream_identifier',
        '_event_values_hash',
        '_parser_chain',
        'data_type']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)

  def testGetAttributes(self):
    """Tests the GetAttributes function."""
    attribute_container = events.EventData()

    with self.assertRaises(TypeError):
      attribute_container.error = b'bytes'
      attribute_container.GetAttributeValuesHash()

    with self.assertRaises(TypeError):
      attribute_container.error = {'key': 'value'}
      attribute_container.GetAttributeValuesHash()

  def testGetEventDataStreamIdentifier(self):
    """Tests the GetEventDataStreamIdentifier function."""
    attribute_container = events.EventData()

    identifier = attribute_container.GetEventDataStreamIdentifier()
    self.assertIsNone(identifier)

  def testSetEventDataStreamIdentifier(self):
    """Tests the SetEventDataStreamIdentifier function."""
    attribute_container = events.EventData()

    attribute_container.SetEventDataStreamIdentifier(None)


class EventDataStreamTest(shared_test_lib.BaseTestCase):
  """Tests for the event data stream attribute container."""

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = events.EventDataStream()

    expected_attribute_names = [
        'file_entropy',
        'md5_hash',
        'path_spec',
        'sha1_hash',
        'sha256_hash',
        'yara_match']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)


class EventObjectTest(shared_test_lib.BaseTestCase):
  """Tests for the event attribute container."""

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = events.EventObject()

    expected_attribute_names = [
        '_event_data_identifier',
        'date_time',
        'timestamp',
        'timestamp_desc']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)

  def testGetEventDataIdentifier(self):
    """Tests the GetEventDataIdentifier function."""
    attribute_container = events.EventObject()

    identifier = attribute_container.GetEventDataIdentifier()
    self.assertIsNone(identifier)

  def testSetEventDataIdentifier(self):
    """Tests the SetEventDataIdentifier function."""
    attribute_container = events.EventObject()

    attribute_container.SetEventDataIdentifier(None)


class EventTagTest(shared_test_lib.BaseTestCase):
  """Tests for the event tag attribute container."""

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = events.EventTag()

    expected_attribute_names = [
        '_event_identifier',
        'labels']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)

  def testGetEventIdentifier(self):
    """Tests the GetEventIdentifier function."""
    attribute_container = events.EventTag()

    identifier = attribute_container.GetEventIdentifier()
    self.assertIsNone(identifier)

  def testSetEventIdentifier(self):
    """Tests the SetEventIdentifier function."""
    attribute_container = events.EventTag()

    attribute_container.SetEventIdentifier(None)


class YearLessLogHelperTest(shared_test_lib.BaseTestCase):
  """Tests for the year-less log helper attribute container."""

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = events.YearLessLogHelper()

    expected_attribute_names = [
        '_event_data_stream_identifier',
        'earliest_year',
        'last_relative_year',
        'latest_year']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)

  def testGetEventDataStreamIdentifier(self):
    """Tests the GetEventDataStreamIdentifier function."""
    attribute_container = events.YearLessLogHelper()

    identifier = attribute_container.GetEventDataStreamIdentifier()
    self.assertIsNone(identifier)

  def testSetEventDataStreamIdentifier(self):
    """Tests the SetEventDataStreamIdentifier function."""
    attribute_container = events.YearLessLogHelper()

    attribute_container.SetEventDataStreamIdentifier(None)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Mozilla Firefox history database plugin."""

import unittest

from plaso.parsers.sqlite_plugins import firefox_history

from tests.parsers.sqlite_plugins import test_lib


class FirefoxHistoryPluginTest(test_lib.SQLitePluginTestCase):
  """Tests for the Mozilla Firefox history database plugin."""

  def testProcessPriorTo24(self):
    """Tests the Process function on a Firefox History database file."""
    # This is probably version 23 but potentially an older version.
    plugin = firefox_history.FirefoxHistoryPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['places.sqlite'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 103)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # Check a page visited entry.
    expected_event_values = {
        'data_type': 'firefox:places:page_visited',
        'host': 'news.google.com',
        'last_visited_time': '2011-07-01T11:16:21.371935+00:00',
        'title': 'Google News',
        'url': 'http://news.google.com/',
        'visit_count': 1,
        'visit_type': 2}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    # Check a bookmark entry.
    expected_event_values = {
        'added_time': '2011-07-01T11:13:59.266344+00:00',
        'data_type': 'firefox:places:bookmark',
        'modification_time': '2011-07-01T11:13:59.267198+00:00',
        'places_title': (
            'folder=BOOKMARKS_MENU&folder=UNFILED_BOOKMARKS&folder=TOOLBAR&'
            'sort=12&excludeQueries=1&excludeItemIfParentHasAnnotation=livemark'
            '%2FfeedURI&maxResults=10&queryType=1'),
        'type': 'URL',
        'url': (
            'place:folder=BOOKMARKS_MENU&folder=UNFILED_BOOKMARKS&folder='
            'TOOLBAR&sort=12&excludeQueries=1&excludeItemIfParentHasAnnotation='
            'livemark%2FfeedURI&maxResults=10&queryType=1'),
        'visit_count': 0}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)

    # Check a bookmark annotation entry.
    expected_event_values = {
        'added_time': '2011-07-01T11:13:59.267605+00:00',
        'content': 'RecentTags',
        'data_type': 'firefox:places:bookmark_annotation',
        'modification_time': None,
        'title': 'Recent Tags',
        'url': 'place:sort=14&type=6&maxResults=10&queryType=1'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 93)
    self.CheckEventData(event_data, expected_event_values)

    # Check a bookmark folder entry.
    expected_event_values = {
        'added_time': '2011-03-21T10:05:01.553774+00:00',
        'data_type': 'firefox:places:bookmark_folder',
        'modification_time': '2011-07-01T11:14:11.766851+00:00',
        'title': 'Latest Headlines'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 102)
    self.CheckEventData(event_data, expected_event_values)

  def testProcessVersion25(self):
    """Tests the Process function on a Firefox History database file v 25."""
    plugin = firefox_history.FirefoxHistoryPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['places_new.sqlite'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 59)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'firefox:places:page_visited',
        'last_visited_time': '2013-10-30T21:57:11.281942+00:00',
        'host': 'code.google.com',
        'url': 'http://code.google.com/p/plaso',
        'visit_count': 1,
        'visit_type': 2}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 10)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for UTMPX file parser."""

import unittest

from plaso.parsers import utmpx

from tests.parsers import test_lib


class UtmpxParserTest(test_lib.ParserTestCase):
  """Tests for utmpx file parser."""

  def testParse(self):
    """Tests the Parse function."""
    parser = utmpx.UtmpxParser()
    storage_writer = self._ParseFile(['utmpx_mac'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 6)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'macos:utmpx:entry',
        'hostname': 'localhost',
        'pid': 67,
        'terminal': 'console',
        'terminal_identifier': 65583,
        'type': 7,
        'username': 'moxilo',
        'written_time': '2013-11-13T17:52:41.736713+00:00'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Parser for OneDrive Log (ODL/ODLGZ) files.

Reference: https://forensics.wiki/microsoft_onedrive/
"""

import base64
import io
import json
import os
import re
import zlib

import pycaes

from dfdatetime import posix_time as dfdatetime_posix_time

from dfvfs.helpers import text_file as dfvfs_text_file

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager


class OneDriveLogEvent(events.EventData):
  """OneDrive log event data.

  Attributes:
    code_filename (str): code filename.
    code_function_name (str): code function name.
    decoded_parameters (str): decoded (and decrypted) parameters.
    raw_parameters (str): the raw parameters encoded as a hexadecimal
        formatted string.
    recorded_time (dfdatetime.DateTimeValues): date and time the entry was
        recorded.
  """

  DATA_TYPE = 'windows:onedrive:log'

  def __init__(self):
    """Initializes event data."""
    super(OneDriveLogEvent, self).__init__(data_type=self.DATA_TYPE)
    self.code_filename = None
    self.code_function_name = None
    self.decoded_parameters = None
    self.raw_parameters = None
    self.recorded_time = None


class OneDriveLogFileParser(
    interface.FileEntryParser, dtfabric_helper.DtFabricHelper):
  """Parser for OneDrive log files."""

  NAME = 'onedrive_log'
  DATA_FORMAT = 'OneDrive Log file'

  BLOCK_SIGNATURE = b'\xcc\xdd\xee\xff\x00\x00\x00\x00'
  COMPRESSED_BLOCK_SIGNATURE = b'\x1f\x8b\x08\x00\x00\x00\x00\x00'

  _AES_IV = b'\x00' * 16
  _AES_KEY_LENGTH = 32
  _BASE64_ALPHABET = (
      'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-')
  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'onedrive.yaml')
  _DELIMITERS = [r'%', r'\\', r'/', r'://', r'\?', r':', r'\.']
  _DELIMITERS_REGEX_PATTERN = '({0:s})'.format('|'.join(_DELIMITERS))
  _DELIMITERS_REGEX = re.compile(_DELIMITERS_REGEX_PATTERN)
  _FIRST_DATA_BLOCK_OFFSET = 256
  _GENERAL_KEYSTORE_FILE_NAME = 'general.keystore'
  _MAXIMUM_FILE_SIZE = 5 * 1024 * 1024
  _OBFUSCATED_STRING_MAP_FILE_NAME = 'ObfuscationStringMap.txt'

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'EBFGONED', offset=0)
    return format_specification

  def _ParseObfuscatedStringMap(
      self, obfuscated_string_map_entry):
    """Parses the ObfuscatedStringMap.txt file.

    ObfuscatedStringMap.txt is a UTF-16 little-endian encoded text file where
    each line contains a tab delimited mapping between an obfuscated string
    value (as read from a OneDrive log file) and its original string value.

    Args:
      obfuscated_string_map_entry (dfvfs.FileEntry): the file entry of the
          ObfuscatedStringMap.txt file.

    Returns:
      dict[str, str]: a mapping of an obfuscated string value to its
          original string value.
    """
    file_object = obfuscated_string_map_entry.GetFileObject()

    obfuscated_string_map = {}
    with dfvfs_text_file.TextFile(
        file_object, encoding='utf-16-le') as obfuscated_text_file:
      for line in obfuscated_text_file.readlines():
        line_tokens = line.rstrip().split('\t')
        if len(line_tokens) == 2:
          obfuscated_string, original_string = line_tokens
          if obfuscated_string not in obfuscated_string_map:
            obfuscated_string_map[obfuscated_string] = original_string

    return obfuscated_string_map

  def _ParseGeneralKeyStore(self, file_entry, parser_mediator):
    """Parses the general.keystore file.

    Args:
      file_entry (dfvfs.FileEntry): the file entry of the general.keystore file.
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.

    Returns:
      Optional[bytes]: the parsed and base64-decoded AES key or None if the
          general.keystore file could not be parsed as JSON or does not contain
          a valid Key value.
    """
    file_object = file_entry.GetFileObject()

    # Note that _MAXIMUM_FILE_SIZE prevents this read to become too large.
    general_keystore_data = file_object.read()

    json_data = json.loads(general_keystore_data)
    if not json_data or not isinstance(json_data, list):
      parser_mediator.ProduceExtractionWarning(
          'Unable to parse OneDrive keystore file: {0:s}'.format(
              file_entry.name))
      return None

    key = json_data[0].get('Key', None)
    if not key:
      parser_mediator.ProduceExtractionWarning(
          'OneDrive keystore file: {0:s} does not contain a key value.'.format(
              file_entry.name))
      return None

    decoded_key = base64.b64decode(key)

    if len(decoded_key) != self._AES_KEY_LENGTH:
      parser_mediator.ProduceExtractionWarning((
          'OneDrive keystore file: {0:s} does not contain a key value of '
          'size: 32.').format(file_entry.name))
      return None

    return decoded_key

  def _DeobfuscateStrings(self, extracted_strings, obfuscated_string_map):
    """Deobfuscate strings extracted from OneDrive logs.

    Args:
      extracted_strings (list[str]): strings extracted from OneDrive log files.
      obfuscated_string_map (dict[str, str]): the parsed obfuscated string map.

    Returns:
      list[str]: the list of extracted strings, replaced by their true value if
          it exists in the obfuscated string map.
    """
    if not obfuscated_string_map:
      return extracted_strings

    deobfuscated_strings = []

    for extracted_string in extracted_strings:
      extracted_string_tokens = self._DELIMITERS_REGEX.split(extracted_string)
      output_string_parts = []

      for extracted_string_token in extracted_string_tokens:
        if extracted_string_token in self._DELIMITERS:
          output_string_parts.append(extracted_string_token)
        else:
          deobfuscated_string_token = obfuscated_string_map.get(
              extracted_string_token, extracted_string_token)
          output_string_parts.append(deobfuscated_string_token)

      deobfuscated_strings.append(''.join(output_string_parts))

    return deobfuscated_strings

  def _DecryptAES(self, aes_key, encrypted_data):
    """Decrypt AES-CBC encrypted data.

    Args:
      aes_key (bytes): the AES key.
      data (bytes): AES-CBC encrypted data.

    Returns:
      bytes: decrypted data.
    """
    aes_context = pycaes.context()
    aes_context.set_key(pycaes.crypt_modes.DECRYPT, aes_key)
    return pycaes.crypt_cbc(
        aes_context, pycaes.crypt_modes.DECRYPT,
        self._AES_IV, encrypted_data)

  def _DecryptStrings(self, parser_mediator, aes_key, extracted_strings):
    """Decrypt strings extracted from OneDrive logs.

    The encrypted strings are stored as follows:
    * split into tokens using delimiters ('%', '\', '/', '://', '?', ':', '.')
    * URL safe base64 encoded
    * encrypted using AES-CBC where the AES key is parsed from the
      general.keystore file and the IV is 16 zero byte values
    * padded using PKCS#7
    * encoded in UTF-16 little-endian

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      aes_key (bytes): the AES key.
      extracted_strings (list[str]): strings extracted from OneDrive log files.

    Returns:
      list[str]: the decrypted strings.
    """
    decrypted_strings = []

    for extracted_string in extracted_strings:
      tokens = self._DELIMITERS_REGEX.split(extracted_string)

      output_string_parts = []
      for token in tokens:
        if token in self._DELIMITERS:
          output_string_parts.append(token)
          continue

        token_size = len(token)

        # The minimum block size of base64 encoded AES-CBC is 22 bytes.
        if token_size < 22:
          output_string_parts.append(token)
          continue

        if not all(character in self._BASE64_ALPHABET for character in token):
          output_string_parts.append(token)
          continue

        base64_padding_size = token_size % 4
        if base64_padding_size == 1:
          output_string_parts.append(token)
          continue

        if base64_padding_size > 1:
          base64_padding = '=' * base64_padding_size
          token = ''.join([token, base64_padding])

        ciphertext = base64.urlsafe_b64decode(token)

        if len(ciphertext) % 16:
          output_string_parts.append(token)
          continue

        plaintext = self._DecryptAES(aes_key, ciphertext)

        try:
          data = self._RemovePKCS7Padding(plaintext, block_size=128)
        except ValueError as exception:
          parser_mediator.ProduceExtractionWarning(
              'unable to remove PKCS#7 padding with error: {0!s}'.format(
                  exception))

          output_string_parts.append(token)
          continue

        try:
          decoded_string = data.decode('utf-16-le')
        except UnicodeError as exception:
          parser_mediator.ProduceExtractionWarning(
              'unable to decode string with error: {0!s}'.format(exception))

          output_string_parts.append(token)
          continue

        output_string_parts.append(decoded_string)

      decrypted_strings.append(''.join(output_string_parts))

    return decrypted_strings

  def _ExtractStringsFromParameters(self, buffer):
    """Extracts strings from raw parameter bytes.

    Args:
      buffer (bytearray): the buffer.

    Returns:
      list[str]: strings parsed from the raw parameter bytes.
    """
    pascal_string_map = self._GetDataTypeMap('pascal_string')

    parse_offset = 0
    strings_array = []

    while parse_offset < len(buffer):
      try:
        pascal_string = self._ReadStructureFromByteStream(
            buffer[parse_offset:], 0, pascal_string_map)
      except errors.ParseError:
        parse_offset += 4
        continue

      strings_array.append(pascal_string.value)
      parse_offset += 4 + pascal_string.size

    return strings_array

  def _ProcessRawParameters(
      self, parser_mediator, raw_parameters_data, aes_key,
      obfuscated_string_map):
    """Processes (extracts and decrypts/deobfuscates) the raw parameters.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      raw_parameters_data (bytes): the raw parameters data from a log entry.
      aes_key (bytes): the AES key used to decrypt strings, or None if there
          is no parsed AES key.
      obfuscated_string_map (dict[str, str]): the obfuscated string map used
          to deobfuscate strings, or None if there is no parsed string mapping.

    Returns:
      list[str]: a list of strings that were parsed from the raw parameters.
    """
    extracted_strings = self._ExtractStringsFromParameters(raw_parameters_data)
    if aes_key:
      extracted_strings = self._DecryptStrings(
          parser_mediator, aes_key, extracted_strings)

    if obfuscated_string_map:
      extracted_strings = self._DeobfuscateStrings(
          extracted_strings, obfuscated_string_map)

    return extracted_strings

  def _RemovePKCS7Padding(self, padded_data, block_size=8):
    """Validates and removes PKCS#7 padding.

    Args:
      padded_data (bytes): data with PKCS#7 padding.
      block_size (Optional[int]): block size.

    Returns:
      bytes: data without PKCS#7 padding.

    Raises:
      ValueError: if the padding is invalid.
    """
    if padded_data:
      padding_size = padded_data[-1]
    else:
      padding_size = 0

    if padding_size <= 0 or padding_size > block_size:
      raise ValueError(
          'Invalid padding size: {0:d} value out of bounds'.format(
              padding_size))

    for byte_index in range(2, padding_size + 1):
      check_padding_size = padded_data[-byte_index]
      if check_padding_size != padding_size:
        raise ValueError((
            'Mismatch in padding size: {0:d} and: {1:d} at byte index: '
            '{2:d}').format(padding_size, check_padding_size, -byte_index))

    return padded_data[:-padding_size]

  def ParseFileEntry(self, parser_mediator, file_entry):
    """Parses a OneDrive Log file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_entry (dfvfs.FileEntry): file entry.

    Raises:
      ParseError: when a log cannot be decompressed.
      WrongParser: when the file cannot be parsed.
    """
    onedrivelog_file_header_map = self._GetDataTypeMap(
        'onedrivelog_file_header')

    file_object = file_entry.GetFileObject()

    try:
      _, data_size = self._ReadStructureFromFileObject(
          file_object, 0, onedrivelog_file_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.WrongParser(
          'Unable to parse OneDrive Log file header with error: {0!s}'.format(
              exception))

    # skip to the first data block offset and check the signature
    file_object.seek(self._FIRST_DATA_BLOCK_OFFSET, os.SEEK_SET)
    data_block_signature = file_object.read(8)

    if data_block_signature == self.COMPRESSED_BLOCK_SIGNATURE:
      file_object.seek(self._FIRST_DATA_BLOCK_OFFSET, os.SEEK_SET)

      # Note that _MAXIMUM_FILE_SIZE prevents this read to become too large.
      raw_data_block = file_object.read()
      try:
        raw_data_block = zlib.decompress(
            raw_data_block, wbits = zlib.MAX_WBITS | 16)
      except zlib.error as zlib_error:
        raise errors.ParseError(
            'Error decompressing data block: {0!s}'.format(str(zlib_error)))

      data_block_stream = io.BytesIO(raw_data_block)
      data_block_size = len(data_block_stream.getbuffer())
      stream_offset = 0

    elif data_block_signature == self.BLOCK_SIGNATURE:
      file_object.seek(self._FIRST_DATA_BLOCK_OFFSET, os.SEEK_SET)
      data_block_stream = file_object
      data_block_size = file_entry.size
      stream_offset = 256

    else:
      raise errors.ParseError('Invalid signature found.')

    parent_file_entry = file_entry.GetParentFileEntry()

    obfuscated_string_map_file_entry = parent_file_entry.GetSubFileEntryByName(
        self._OBFUSCATED_STRING_MAP_FILE_NAME)
    if obfuscated_string_map_file_entry:
      obfuscated_string_map = self._ParseObfuscatedStringMap(
          obfuscated_string_map_file_entry)
    else:
      obfuscated_string_map = None

    aes_key_file_entry = parent_file_entry.GetSubFileEntryByName(
        self._GENERAL_KEYSTORE_FILE_NAME)
    if aes_key_file_entry:
      aes_key = self._ParseGeneralKeyStore(aes_key_file_entry, parser_mediator)
    else:
      aes_key = None

    odl_data_block_header_map = self._GetDataTypeMap('odl_data_block_header')
    odl_data_block_map = self._GetDataTypeMap('odl_data_block')

    while (stream_offset + odl_data_block_header_map.GetSizeHint() <
        data_block_size):
      try:
        odl_data_block_header, header_data_size = (
            self._ReadStructureFromFileObject(
                data_block_stream, stream_offset, odl_data_block_header_map))
      except (ValueError, errors.ParseError) as exception:
        parser_mediator.ProduceExtractionWarning((
            'unable to parse data block header at stream offset: 0x{0:08x} '
            'with error: {1!s}').format(stream_offset, exception))
        return

      stream_offset += header_data_size

      try:
        odl_data_block, data_size = self._ReadStructureFromFileObject(
            data_block_stream, stream_offset, odl_data_block_map)
      except (ValueError, errors.ParseError) as exception:
        parser_mediator.ProduceExtractionWarning((
            'unable to parse data block at stream offset: 0x{0:08x} with '
            'error: {1!s}').format(stream_offset, exception))
        return

      # read the raw function parameter data
      stream_offset += data_size
      raw_parameters_size = odl_data_block_header.data_size - data_size
      raw_parameters_data = data_block_stream.read(raw_parameters_size)
      stream_offset += raw_parameters_size

      extracted_strings = self._ProcessRawParameters(
          parser_mediator, raw_parameters_data, aes_key, obfuscated_string_map)

      event_data = OneDriveLogEvent()
      event_data.code_filename = odl_data_block.code_filename.decode('utf-8')
      event_data.code_function_name = odl_data_block.code_function_name.decode(
          'utf-8')

      event_data.decoded_parameters = extracted_strings
      event_data.raw_parameters = raw_parameters_data.hex()
      event_data.recorded_time = dfdatetime_posix_time.PosixTimeInMilliseconds(
          timestamp=odl_data_block_header.timestamp)
      parser_mediator.ProduceEventData(event_data)


manager.ParsersManager.RegisterParser(OneDriveLogFileParser)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the log2timeline CLI tool."""

import collections
import os
import platform
import unittest

from plaso.containers import events
from plaso.cli import log2timeline_tool
from plaso.lib import definitions
from plaso.lib import errors
from plaso.storage.sqlite import sqlite_file

from tests import test_lib as shared_test_lib
from tests.cli import test_lib


class Log2TimelineToolTest(test_lib.CLIToolTestCase):
  """Tests for the log2timeline CLI tool."""

  # pylint: disable=protected-access

  _BDE_PASSWORD = 'bde-TEST'
  _OUTPUT_ENCODING = 'utf-8'

  def _CheckOutput(self, output, expected_output):
    """Compares the output against the expected output.

    The actual processing time is ignored, since it can vary.

    Args:
      output (str): tool output.
      expected_output (list[str]): expected tool output.
    """
    output = output.split('\n')

    self.assertEqual(output[:3], expected_output[:3])
    self.assertTrue(output[3].startswith('Processing time\t\t: '))
    self.assertEqual(output[4:], expected_output[4:])

  def _CreateExtractionOptions(self, source_path, password=None):
    """Create options for testing extraction.

    Args:
      source_path (str): path of the source (test) data.
      password (Optional[str]): password to unlock test data.

    Returns:
      TestOptions: options for testing extraction.
    """
    options = test_lib.TestOptions()
    options.artifact_definitions_path = self._GetTestFilePath(['artifacts'])
    options.data_location = shared_test_lib.DATA_PATH
    options.quiet = True
    options.single_process = True
    options.status_view_interval = 0.5
    options.status_view_mode = 'none'
    options.source = source_path

    if password:
      options.credentials = ['password:{0:s}'.format(password)]

    return options

  # TODO: add tests for _CheckStorageFile
  # TODO: add tests for _CreateProcessingConfiguration

  def testGetPluginData(self):
    """Tests the _GetPluginData function."""
    test_tool = log2timeline_tool.Log2TimelineTool()
    test_tool._data_location = self._GetTestFilePath([])

    plugin_info = test_tool._GetPluginData()

    self.assertIn('Hashers', plugin_info)

    available_hasher_names = [name for name, _ in plugin_info['Hashers']]
    self.assertIn('sha256', available_hasher_names)
    self.assertIn('sha1', available_hasher_names)

    self.assertIn('Parsers', plugin_info)
    self.assertIsNotNone(plugin_info['Parsers'])

    self.assertIn('Parser Plugins', plugin_info)
    self.assertIsNotNone(plugin_info['Parser Plugins'])

  def CheckEventCounters(self, storage_file, expected_event_counters):
    """Asserts that the number of events per data type matches.

    Args:
      storage_file (StorageFile): storage file.
      expected_event_counters (dict[str, int|list[int]]): expected event
          counters per event data type.
    """
    event_counters = collections.Counter()
    for event in storage_file.GetSortedEvents():
      event_data_identifier = event.GetEventDataIdentifier()
      event_data = storage_file.GetAttributeContainerByIdentifier(
          events.EventData.CONTAINER_TYPE, event_data_identifier)

      event_counters[event_data.data_type] += 1

    for data_type, expected_event_count in expected_event_counters.items():
      event_count = event_counters.pop(data_type, 0)
      if isinstance(expected_event_count, list):
        self.assertIn(event_count, expected_event_count)
      else:
        self.assertEqual(event_count, expected_event_count)

    # Ensure there are no events left unaccounted for.
    self.assertEqual(event_counters, collections.Counter())

  # TODO: add tests for _PrintProcessingSummary

  def testParseArguments(self):
    """Tests the ParseArguments function."""
    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    result = test_tool.ParseArguments([])
    self.assertFalse(result)

    # TODO: check output.
    # TODO: improve test coverage.

  def testParseOptions(self):
    """Tests the ParseOptions function."""
    test_artifacts_path = self._GetTestFilePath(['artifacts'])
    self._SkipIfPathNotExists(test_artifacts_path)

    test_file_path = self._GetTestFilePath(['testdir'])
    self._SkipIfPathNotExists(test_file_path)

    yara_rules_path = self._GetTestFilePath(['rules.yara'])
    self._SkipIfPathNotExists(yara_rules_path)

    options = test_lib.TestOptions()
    options.artifact_definitions_path = test_artifacts_path
    options.source = test_file_path
    options.status_view_interval = 0.5
    options.status_view_mode = 'none'
    options.storage_file = 'storage.plaso'
    options.storage_format = definitions.STORAGE_FORMAT_SQLITE
    options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE
    options.yara_rules_path = yara_rules_path

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)
    test_tool.ParseOptions(options)

    self.assertIsNotNone(test_tool._yara_rules_string)

    options = test_lib.TestOptions()
    options.artifact_definitions_path = test_artifacts_path

    # ParseOptions will raise if source is not set.
    with self.assertRaises(errors.BadConfigOption):
      test_tool.ParseOptions(options)

    options = test_lib.TestOptions()
    options.artifact_definitions_path = test_artifacts_path
    options.source = test_file_path

    with self.assertRaises(errors.BadConfigOption):
      test_tool.ParseOptions(options)

    # TODO: improve test coverage.

  def testExtractEventsFromSourcesOnDirectory(self):
    """Tests the ExtractEventsFromSources function on a directory."""
    test_file_path = self._GetTestFilePath(['testdir'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: directory',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesOnAPFSImage(self):
    """Tests the ExtractEventsFromSources function on APFS image."""
    test_file_path = self._GetTestFilePath(['apfs.dmg'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: storage media image',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesOnBDEImage(self):
    """Tests the ExtractEventsFromSources function on BDE image."""
    test_file_path = self._GetTestFilePath(['bdetogo.raw'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(
        test_file_path, password=self._BDE_PASSWORD)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: storage media image',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesOnCompressedDMGImage(self):
    """Tests the ExtractEventsFromSources function on a compressed DMG image."""
    test_file_path = self._GetTestFilePath(['hfsplus_zlib.dmg'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: storage media image',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesImage(self):
    """Tests the ExtractEventsFromSources function on single partition image."""
    test_file_path = self._GetTestFilePath(['ímynd.dd'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: storage media image',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesPartitionedImage(self):
    """Tests the ExtractEventsFromSources function on multi partition image."""
    # Note that the source file is a RAW (VMDK flat) image.
    test_file_path = self._GetTestFilePath(['multi_partition_image.vmdk'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)
    options.partitions = 'all'

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: storage media image',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesOnVSSImage(self):
    """Tests the ExtractEventsFromSources function on VSS image."""
    test_file_path = self._GetTestFilePath(['vsstest.qcow2'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)
    options.unattended = True
    options.vss_stores = 'all'

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: storage media image',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          'Number of warnings generated while extracting events: 3.',
          '',
          'Use pinfo to inspect warnings in more detail.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesOnFile(self):
    """Tests the ExtractEventsFromSources function on a file."""
    test_file_path = self._GetTestFilePath(['System.evtx'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: single file',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  @unittest.skipIf(platform.system() == 'Windows', 'not supported on Windows')
  @unittest.skipIf(
      platform.release().endswith('Microsoft'),
      'not supported on Windows Subsystem for Linux')
  def testExtractEventsFromSourcesOnLinkToDirectory(self):
    """Tests the ExtractEventsFromSources function on a symlink to directory."""
    test_file_path = self._GetTestFilePath(['link_to_testdir'])
    self._SkipIfPathNotExists(test_file_path)

    options = self._CreateExtractionOptions(test_file_path)

    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: directory',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  @unittest.skipIf(platform.system() == 'Windows', 'not supported on Windows')
  @unittest.skipIf(
      platform.release().endswith('Microsoft'),
      'not supported on Windows Subsystem for Linux')
  def testExtractEventsFromSourcesOnLinkToFile(self):
    """Tests the ExtractEventsFromSources function on a symlink to file."""
    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    source_path = self._GetTestFilePath(['link_to_System.evtx'])
    options = self._CreateExtractionOptions(source_path)

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      expected_output = [
          '',
          'Source path\t\t: {0:s}'.format(options.source),
          'Source type\t\t: single file',
          'Processing time\t\t: 00:00:00',
          '',
          'Processing started.',
          'Processing completed.',
          '',
          '']

      output = output_writer.ReadOutput()
      self._CheckOutput(output, expected_output)

  def testExtractEventsFromSourcesWithFilestat(self):
    """Tests the ExtractEventsFromSources function with filestat parser."""
    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    source_path = self._GetTestFilePath(['test_pe.exe'])
    options = self._CreateExtractionOptions(source_path)
    options.parsers = 'filestat,pe'

    with shared_test_lib.TempDirectory() as temp_directory:
      options.storage_file = os.path.join(temp_directory, 'storage.plaso')
      options.storage_format = definitions.STORAGE_FORMAT_SQLITE
      options.task_storage_format = definitions.STORAGE_FORMAT_SQLITE

      test_tool.ParseOptions(options)

      test_tool.ExtractEventsFromSources()

      storage_file = sqlite_file.SQLiteStorageFile()
      try:
        storage_file.Open(path=options.storage_file, read_only=True)
      except IOError as exception:
        self.fail((
            'Unable to open storage file after processing with error: '
            '{0!s}.').format(exception))

      # There should be 3 filestat and 3 pe parser generated events.
      # Typically there are 3 filestat events, but there can be 4 on platforms
      # that support os.stat_result st_birthtime.
      expected_event_counters = {
          'fs:stat': [3, 4],
          'pe_coff:dll_import': 2,
          'pe_coff:file': 1}

      self.CheckEventCounters(storage_file, expected_event_counters)

  def testShowInfo(self):
    """Tests the output of the tool in info mode."""
    output_writer = test_lib.TestOutputWriter(encoding=self._OUTPUT_ENCODING)
    test_tool = log2timeline_tool.Log2TimelineTool(output_writer=output_writer)

    options = test_lib.TestOptions()
    options.artifact_definitions_path = self._GetTestFilePath(['artifacts'])
    options.show_info = True

    test_tool.ParseOptions(options)
    test_tool.ShowInfo()

    output = output_writer.ReadOutput()

    section_headings = [
        'Hashers', 'Parsers', 'Parser Plugins', 'Parser Presets',
        'Versions']
    for heading in section_headings:
      self.assertIn(heading, output)

    self.assertNotIn('<class', output)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Windows Scheduled Task job file parser."""

import unittest

from plaso.parsers import winjob

from tests.parsers import test_lib


class WinJobTest(test_lib.ParserTestCase):
  """Tests for the Windows Scheduled Task job file parser."""

  def testParse(self):
    """Tests the Parse function."""
    parser = winjob.WinJobParser()
    storage_writer = self._ParseFile(['wintask.job'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 2)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'application': (
            'C:\\Program Files (x86)\\Google\\Update\\GoogleUpdate.exe'),
        'comment': (
            'Keeps your Google software up to date. If this task is disabled '
            'or stopped, your Google software will not be kept up to date, '
            'meaning security vulnerabilities that may arise cannot be fixed '
            'and features may not work. This task uninstalls itself when there '
            'is no Google software using it.'),
        'data_type': 'windows:tasks:job',
        'parameters': '/ua /installsource scheduler',
        'last_run_time': '2013-08-24T12:42:00.112+00:00',
        'username': 'Brian',
        'working_directory': None}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'application': (
            'C:\\Program Files (x86)\\Google\\Update\\GoogleUpdate.exe'),
        'comment': (
            'Keeps your Google software up to date. If this task is disabled '
            'or stopped, your Google software will not be kept up to date, '
            'meaning security vulnerabilities that may arise cannot be fixed '
            'and features may not work. This task uninstalls itself when there '
            'is no Google software using it.'),
        'data_type': 'windows:tasks:trigger',
        'end_time': None,
        'parameters': '/ua /installsource scheduler',
        'start_time': '2013-07-12T15:42:00',
        'trigger_type': 1,
        'username': 'Brian',
        'working_directory': None}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Analysis plugin to look up file hashes in bloom database."""

import flor

from plaso.analysis import hash_tagging
from plaso.analysis import logger
from plaso.analysis import manager


class BloomAnalysisPlugin(hash_tagging.HashTaggingAnalysisPlugin):
  """Analysis plugin for looking up hashes in bloom database."""

  DATA_TYPES = frozenset(['fs:stat', 'fs:stat:ntfs'])

  NAME = 'bloom'

  SUPPORTED_HASHES = frozenset(['md5', 'sha1', 'sha256'])

  DEFAULT_LABEL = 'bloom_present'

  def __init__(self):
    """Initializes a bloom database analysis plugin."""
    super(BloomAnalysisPlugin, self).__init__()
    self._bloom_database_path = None
    self._bloom_filter_object = None
    self._label = self.DEFAULT_LABEL

  def _Analyze(self, hashes):
    """Looks up file hashes in a bloom database.

    Args:
      hashes (list[str]): hash values to look up.

    Returns:
      list[HashAnalysis]: analysis results, or an empty list on error.

    Raises:
      RuntimeError: when the analyzer fail to get a bloom filter object.
    """
    bloom_filter = self._GetBloomFilterObject(cached=True)
    if not bloom_filter:
      raise RuntimeError('Failed to open bloom file')

    hash_analyses = []
    for digest in hashes:
      response = self._QueryHash(digest=digest, bloom_filter=bloom_filter)
      if response is not None:
        hash_analysis = hash_tagging.HashAnalysis(
          subject_hash=digest, hash_information=response)
        hash_analyses.append(hash_analysis)

    return hash_analyses

  def _GenerateLabels(self, hash_information):
    """Generates a list of strings that will be used in the event tag.

    Args:
      hash_information (bool): response from the hash tagging that indicates
          that the file hash was present or not.

    Returns:
      list[str]: list of labels to apply to event.
    """
    if hash_information:
      return [self._label]
    return []

  def _GetBloomFilterObject(self, cached=True):
    """Loads a bloom filter file in memory.

    Args:
      cached (bool): True if the bloom filter should be cached.

    Returns:
      flor.BloomFilter: bloom filter object or None if not available.
    """
    if self._bloom_filter_object:
      return self._bloom_filter_object

    logger.debug('Open bloom database file: {0:s}.'.format(
        self._bloom_database_path))

    if flor.BloomFilter is None:
      logger.error('Missing optional dependency : flor')
      return None

    try:
      bloom_filter = flor.BloomFilter()
      with open(self._bloom_database_path, 'rb') as file_object:
        bloom_filter.read(file_object)

    except IOError as exception:
      bloom_filter = None
      logger.error(('Unable to open bloom database file: {0:s} with error: '
                    '{1!s}.').format(self._bloom_database_path, exception))

    if cached:
      self._bloom_filter_object = bloom_filter
      return self._bloom_filter_object
    return bloom_filter

  def _QueryHash(self, digest, bloom_filter):
    """Queries BloomFilter for a specific hash in upper case.

    Args:
      digest (str): hash to look up.
      bloom_filter (flor.BloomFilter): instanced bloom filter.

    Returns:
      bool: True if the hash was found, False if not.
    """
    value_to_test = digest.upper().encode()
    return value_to_test in bloom_filter

  def SetBloomDatabasePath(self, bloom_database_path):
    """Set the path to the bloom file containing hash

    Args:
      bloom_database_path (str): Path to the bloom file
    """
    self._bloom_database_path = bloom_database_path

  def SetLabel(self, label):
    """Sets the tagging label.

    Args:
      label (str): label to apply to events extracted from files that are
          present in the bloom database.
    """
    self._label = label

  def TestLoading(self):
    """Checks if the bloom database exist and is valid.

    Returns:
      bool: True is the bloom database exist and is valid.
    """
    return bool(self._GetBloomFilterObject(cached=False))


manager.AnalysisPluginManager.RegisterPlugin(BloomAnalysisPlugin)



# -*- coding: utf-8 -*-
"""SQLite parser plugin for iOS netusage.sqlite database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class IOSNetusageProcessEventData(events.EventData):
  """iOS netusage process event data.

  Attributes:
    process_name (str): name of the process.
    start_time (dfdatetime.DateTimeValues): date and time the start of
        the network connection was established.
    wifi_in (int): bytes received via wifi.
    wifi_out (int): bytes sent via wifi.
    wired_in (int): bytes received via wired connection.
    wired_out (int): bytes sent via wired connection.
    wireless_wan_in (int): bytes received via cellular connection.
    wireless_wan_out (int): bytes sent via cellular connection.
  """
  DATA_TYPE = 'ios:netusage:process'

  def __init__(self):
    """Initializes event data."""
    super(IOSNetusageProcessEventData, self).__init__(data_type=self.DATA_TYPE)
    self.process_name = None
    self.start_time = None
    self.wifi_in = None
    self.wifi_out = None
    self.wired_in = None
    self.wired_out = None
    self.wireless_wan_in = None
    self.wireless_wan_out = None


class IOSNetusageRouteEventData(events.EventData):
  """iOS netusage connection event data.

  Attributes:
    bytes_in (int): number of bytes received.
    bytes_out (int): number of bytes sent.
    network_identifier (str): name of network.
    network_signature (str): signature of network.
    network_type (int): integer indicating network type.
    start_time (dfdatetime.DateTimeValues): date and time the start of
        the network connection was established.
  """

  DATA_TYPE = 'ios:netusage:route'

  def __init__(self):
    """Initializes event data."""
    super(IOSNetusageRouteEventData, self).__init__(data_type=self.DATA_TYPE)
    self.bytes_in = None
    self.bytes_out = None
    self.network_identifier = None
    self.network_signature = None
    self.network_type = None
    self.start_time = None


class IOSNetusagePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for iOS netusage database."""

  NAME = 'ios_netusage'
  DATA_FORMAT = 'iOS network usage SQLite database (netusage.sqlite) file'

  REQUIRED_STRUCTURE = {
      'ZLIVEROUTEPERF': frozenset([
          'ZTIMESTAMP', 'ZKIND', 'ZBYTESIN', 'ZBYTESOUT',
          'ZHASNETWORKATTACHMENT']),
      'ZNETWORKATTACHMENT': frozenset(['Z_PK', 'ZIDENTIFIER', 'ZNETSIGNATURE']),
      'ZLIVEUSAGE': frozenset([
          'ZTIMESTAMP', 'ZWIFIIN', 'ZWIFIOUT', 'ZWWANIN', 'ZWWANOUT',
          'ZWIREDIN', 'ZWIREDOUT', 'ZHASPROCESS']),
      'ZPROCESS': frozenset(['Z_PK', 'ZPROCNAME'])}

  QUERIES = [
      ("""
       SELECT 
         ZLIVEROUTEPERF.ZTIMESTAMP, 
         ZLIVEROUTEPERF.ZKIND, 
         ZLIVEROUTEPERF.ZBYTESIN, 
         ZLIVEROUTEPERF.ZBYTESOUT, 
         ZNETWORKATTACHMENT.ZIDENTIFIER, 
         HEX(ZNETWORKATTACHMENT.ZNETSIGNATURE) as ZNETSIGNATURE
       FROM ZLIVEROUTEPERF
       LEFT JOIN ZNETWORKATTACHMENT 
       ON ZLIVEROUTEPERF.ZHASNETWORKATTACHMENT = ZNETWORKATTACHMENT.Z_PK""",
       'ParseNetusageRouteRow'),
      ("""
         SELECT
           ZLIVEUSAGE.ZTIMESTAMP,
           ZPROCESS.ZPROCNAME,
           ZLIVEUSAGE.ZWIFIIN,
           ZLIVEUSAGE.ZWIFIOUT,
           ZLIVEUSAGE.ZWWANIN,
           ZLIVEUSAGE.ZWWANOUT,
           ZLIVEUSAGE.ZWIREDIN,
           ZLIVEUSAGE.ZWIREDOUT
         FROM ZLIVEUSAGE 
         LEFT JOIN ZPROCESS 
         ON ZPROCESS.Z_PK = ZLIVEUSAGE.ZHASPROCESS""",
      'ParseNetusageProcessRow')
  ]

  SCHEMAS = {
      'ZLIVEROUTEPERF': (
          'CREATE TABLE ZLIVEROUTEPERF (Z_PK INTEGER PRIMARY KEY, '
          'Z_ENT INTEGER, Z_OPT INTEGER, ZKIND INTEGER, ZHASNETWORKATTACHMENT '
          'INTEGER, ZADMINDISABLES FLOAT, ZBYTESIN FLOAT, ZBYTESOUT FLOAT, '
          'ZCAPTIVITYREDIRECTS FLOAT, ZCERTERRORS FLOAT, ZCONNATTEMPTS FLOAT, '
          'ZCONNSUCCESSES FLOAT, ZDATASTALLS FLOAT, ZEPOCHS FLOAT, ZFAULTYSTAY '
          'FLOAT, ZLOWLQMSTAY FLOAT, ZLOWQSTAY FLOAT, ZLQMTRANSITIONCOUNT '
          'FLOAT, ZOVERALLSTAY FLOAT, ZOVERALLSTAYM2 FLOAT, ZPACKETSIN FLOAT, '
          'ZPACKETSOUT FLOAT, ZRETXBYTES FLOAT, ZRTTAVG FLOAT, ZRTTMIN FLOAT, '
          'ZRTTVAR FLOAT, ZRXDUPEBYTES FLOAT, ZRXOOOBYTES FLOAT, ZTIMESTAMP '
          'TIMESTAMP, ZTOPDOWNLOADRATE FLOAT )'),
      'ZNETWORKATTACHMENT': (
          'CREATE TABLE ZNETWORKATTACHMENT ( Z_PK INTEGER PRIMARY KEY, '
          'Z_ENT INTEGER, Z_OPT INTEGER, ZATTRS INTEGER, ZISHOTSPOT INTEGER, '
          'ZISKNOWNGOOD INTEGER, ZISLOWINTERNETDL INTEGER, '
          'ZISLOWINTERNETUL INTEGER, ZKIND INTEGER, ZWASLASTFAILED INTEGER, '
          'ZFIRSTTIMESTAMP TIMESTAMP, ZOVERALLSTAYMEAN FLOAT, '
          'ZOVERALLSTAYVAR FLOAT, ZTIMESTAMP TIMESTAMP, ZVELO FLOAT, '
          'ZIDENTIFIER VARCHAR, ZSERVICE VARCHAR, ZNETSIGNATURE BLOB )'),
      'ZLIVEUSAGE': (
          'CREATE TABLE ZLIVEUSAGE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZKIND INTEGER, ZMETADATA INTEGER, ZTAG INTEGER, '
          'ZHASPROCESS INTEGER, Z15_HASPROCESS INTEGER, ZALLFLOWS FLOAT, '
          'ZBILLCYCLEEND TIMESTAMP, ZJUMBOFLOWS FLOAT, ZTIMESTAMP TIMESTAMP, '
          'ZWIFIIN FLOAT, ZWIFIOUT FLOAT, ZWIREDIN FLOAT, ZWIREDOUT FLOAT, '
          'ZWWANIN FLOAT, ZWWANOUT FLOAT, ZXIN FLOAT, ZXOUT FLOAT )'),
      'ZPROCESS': (
          'CREATE TABLE ZPROCESS ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZFIRSTTIMESTAMP TIMESTAMP, ZTIMESTAMP TIMESTAMP, '
          'ZBUNDLENAME VARCHAR, ZPROCNAME VARCHAR )')}

  REQUIRES_SCHEMA_MATCH = False

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  # pylint: disable=unused-argument
  def ParseNetusageRouteRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a Netusage route row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = IOSNetusageRouteEventData()
    event_data.bytes_in = int(self._GetRowValue(query_hash, row, 'ZBYTESIN'))
    event_data.bytes_out = int(self._GetRowValue(query_hash, row, 'ZBYTESOUT'))
    event_data.network_identifier = self._GetRowValue(
        query_hash, row, 'ZIDENTIFIER')
    event_data.network_signature = self._GetRowValue(
        query_hash, row, 'ZNETSIGNATURE')
    event_data.network_type = self._GetRowValue(query_hash, row, 'ZKIND')
    event_data.start_time = self._GetDateTimeRowValue(
        query_hash, row, 'ZTIMESTAMP')

    parser_mediator.ProduceEventData(event_data)

  # pylint: disable=unused-argument
  def ParseNetusageProcessRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses a Netusage process row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = IOSNetusageProcessEventData()
    event_data.process_name = self._GetRowValue(query_hash, row, 'ZPROCNAME')
    event_data.start_time = self._GetDateTimeRowValue(
        query_hash, row, 'ZTIMESTAMP')
    event_data.wifi_in = int(self._GetRowValue(query_hash, row, 'ZWIFIIN'))
    event_data.wifi_out = int(self._GetRowValue(query_hash, row, 'ZWIFIOUT'))
    event_data.wired_in = int(self._GetRowValue(query_hash, row, 'ZWIREDIN'))
    event_data.wired_out = int(self._GetRowValue(query_hash, row, 'ZWIREDOUT'))
    event_data.wireless_wan_in = int(self._GetRowValue(
        query_hash, row, 'ZWWANIN'))
    event_data.wireless_wan_out = int(self._GetRowValue(
        query_hash, row, 'ZWWANOUT'))

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(IOSNetusagePlugin)
